/* etherdev.c:
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	ed.sutter@alcatel-lucent.com
 * phone:	908-582-2351
 *
 */

#include "config.h"

#if INCLUDE_ETHERNET

#include "stddefs.h"
#include "ether.h"
#include "timer.h"
#include "endian.h"
#include <stdio.h>
#include <string.h>

#include "MK60D10.h"
#include "enet.h"

extern int periph_clk_khz;

#define PHYADDR		1
#define ENETDEV		0

#define RXBUFTOT	4
#define TXBUFTOT	2

#define EnetMaxFrameSize	1518
#define EnetBufSize			1536	//(mod16)
#define ENETTRACE(a)		//printf a 

static uint8_t	lastlinkstate;

int phy_read(int phyReg, __uint16_t *value);
int phy_write(int phyReg, uint16_t value);

typedef struct {
	uint16_t status;
	uint16_t length;
	uint8_t  *data;
#ifdef ENHANCED_BD
	uint32_t ebd_status;
	uint16_t length_proto_type;
	uint16_t payload_checksum;
	uint32_t bdu;
	uint32_t timestamp;
	uint32_t reserverd_word1;
	uint32_t reserverd_word2;
#endif
} enet_bd_struct_t;

/* Buffer descriptors:
 */
static enet_bd_struct_t *rxbdtbl, *txbdtbl;
static uint8_t rxbdbuf[(sizeof(enet_bd_struct_t) * (RXBUFTOT+1))];
static uint8_t txbdbuf[(sizeof(enet_bd_struct_t) * (TXBUFTOT+1))];

/* Buffers:
 */
static uint8_t txbuf[TXBUFTOT][EnetBufSize], rxbuf[RXBUFTOT][EnetBufSize];
static uint8_t *txbuftbl[TXBUFTOT];
static int rxidx, txidx;

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
	struct elapsed_tmr tmr;

	/* Enable ENET clock */
	SIM_SCGC2 |= SIM_SCGC2_ENET_MASK;

	/* Set up pin-muxing... */
	PORTB_PCR0  = PORT_PCR_MUX(4);  //GPIO;//RMII0_MDIO/MII0_MDIO
	PORTB_PCR1  = PORT_PCR_MUX(4);  //GPIO;//RMII0_MDC/MII0_MDC
	PORTA_PCR14 = PORT_PCR_MUX(4);  //RMII0_CRS_DV/MII0_RXDV
	PORTA_PCR12 = PORT_PCR_MUX(4);  //RMII0_RXD1/MII0_RXD1
	PORTA_PCR13 = PORT_PCR_MUX(4);  //RMII0_RXD0/MII0_RXD0
	PORTA_PCR15 = PORT_PCR_MUX(4);  //RMII0_TXEN/MII0_TXEN
	PORTA_PCR16 = PORT_PCR_MUX(4);  //RMII0_TXD0/MII0_TXD0
	PORTA_PCR17 = PORT_PCR_MUX(4);  //RMII0_TXD1/MII0_TXD1

    /* Set the Reset bit and clear the Enable bit */
    ENET_ECR = ENET_ECR_RESET_MASK;

	startElapsedTimer(&tmr,2);
	while(!msecElapsed(&tmr));
}

uint8_t
enet_hash_address(const uint8_t* addr)
{
	int i, j;
	uint8_t byte;
	uint32_t crc;

	crc = 0xFFFFFFFF;
	for(i=0; i<6; ++i) {
		byte = addr[i];
		for(j=0; j<8; ++j) {
			if((byte & 0x01)^(crc & 0x01)) {
				crc >>= 1;
				crc = crc ^ 0xEDB88320;
			}
			else {
				crc >>= 1;
			}
			byte >>= 1;
		}
	}
	return (uint8_t)(crc >> 26);
}

static void
rxbdinit()
{
	int i;
	uint8_t *bdp, *bp;

	/* Initialize the buffer descriptor tables:
	 * (both the buffers and the descriptors have to be 16-byte aligned)
	 */
	memset(rxbdbuf,0,sizeof(rxbdbuf));
	bdp = rxbdbuf;
	while((long)bdp & 0xf)
		bdp++;
	rxbdtbl = (enet_bd_struct_t *)bdp;
	for(i=0;i<RXBUFTOT;i++) {
		bp = rxbuf[i];
		while((long)bp & 0xf)
			bp++;
		ENETTRACE(("&rxbdtbl[%d] = 0x%lx, buf=0x%lx\n",i,(long)&rxbdtbl[i],(long)bp));
		rxbdtbl[i].data = (uint8_t *)ecl((unsigned long)bp);
		rxbdtbl[i].status = RX_BD_E;
		rxbdtbl[i].length = 0;
#ifdef ENHANCED_BD
		rxbdtbl[i].bdu = 0x00000000;
		rxbdtbl[i].ebd_status = RX_BD_INT;
#endif               
	}
	rxbdtbl[RXBUFTOT - 1].status |= RX_BD_W;
	ENET_RDSR = (uint32_t)rxbdtbl;
	rxidx = 0;
}

void
txbdinit(void)
{
	int i;
	uint8_t *bdp, *bp;

	memset(txbdbuf,0,sizeof(txbdbuf));
	bdp = txbdbuf;
	while((long)bdp & 0xf)
		bdp++;
	txbdtbl = (enet_bd_struct_t *)bdp;
	for(i=0;i<TXBUFTOT;i++) {
		bp = txbuf[i];
		while((long)bp & 0xf)
			bp++;
		txbuftbl[i] = bp;
		ENETTRACE(("&txbdtbl[%d] = 0x%lx, buf=0x%lx\n",i,(long)&txbdtbl[i],(long)bp));
		txbdtbl[i].data = (uint8_t *)ecl((unsigned long)bp);
		txbdtbl[i].length = 0;
		txbdtbl[i].status = 0;
#ifdef ENHANCED_BD
		txbdtbl[i].ebd_status = TX_BD_IINS | TX_BD_PINS;
#endif
	}
	txbdtbl[TXBUFTOT - 1].status |= TX_BD_W;
	ENET_TDSR = (uint32_t)txbdtbl;
	txidx = 0;
}

/*
 * eninit():
 * These steps are meant to mimic the SDK function ENET_initialize()
 * which calls enet_mac_init().
 * Return 0 if successful; else -1.
 */

static int
eninit(void)
{
	uint8_t crc;

	/********************
	 * Initialize the MII interface:
	 */
    ENET_MSCR = ENET_MSCR_MII_SPEED((2*(periph_clk_khz/1000)/5)+1);

	/********************
	 * Initialize the MAC:
	 */
    ENET_IALR = 0;
    ENET_IAUR = 0;
    ENET_GALR = 0;
    ENET_GAUR = 0;

    ENET_PALR = (uint32_t)((BinEnetAddr[0]<<24) | (BinEnetAddr[1]<<16) | (BinEnetAddr[2]<<8) | BinEnetAddr[3]);
    ENET_PAUR = (uint32_t)((BinEnetAddr[4]<<24) | (BinEnetAddr[5]<<16));
    /*
     * Calculate and set the hash for given Physical Address
     * in the  Individual Address Hash registers
     */
    crc = enet_hash_address(BinEnetAddr);
    if(crc >= 32)
        ENET_IAUR/*(ch)*/ |= (uint32_t)(1 << (crc - 32));
    else
        ENET_IALR/*(ch)*/ |= (uint32_t)(1 << crc);

    ENET_EIMR = 0;
    ENET_EIR = 0xFFFFFFFF;
    
    /* Initialize the Receive Control Register */
    ENET_RCR = ENET_RCR_MAX_FL(EnetMaxFrameSize) | ENET_RCR_MII_MODE_MASK | ENET_RCR_CRCFWD_MASK;
	ENET_RCR |= ENET_RCR_RMII_MODE_MASK;
    ENET_TCR = 0;    
    
    /* Full duplex... */
	ENET_RCR &= ~ENET_RCR_DRT_MASK;
	ENET_TCR |= ENET_TCR_FDEN_MASK;
        
#ifdef ENHANCED_BD
	ENET_ECR = ENET_ECR_EN1588_MASK;
#else
	ENET_ECR = 0;
#endif

	rxbdinit();
	txbdinit();

	// Set Receive Buffer Size
	ENET_MRBR = (uint16_t)EnetBufSize;

	ENET_EIMR = ENET_EIMR_RXF_MASK | ENET_EIMR_TXF_MASK |
			ENET_EIMR_UN_MASK | ENET_EIMR_RL_MASK | ENET_EIMR_LC_MASK |
			ENET_EIMR_BABT_MASK | ENET_EIMR_BABR_MASK | ENET_EIMR_EBERR_MASK | ENET_EIMR_RXB_MASK;
           
	ENET_ECR |= ENET_ECR_ETHEREN_MASK;
	ENET_RDAR = ENET_RDAR_RDAR_MASK;
	return(0);
}

int
EtherdevStartup(int verbose)
{
	enreset();
	return(eninit());
}

void
disableMulticastReception(void)
{
}

void
enableMulticastReception(void)
{
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	ENET_RCR &= ~ENET_RCR_PROM_MASK; 
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	ENET_RCR |= ENET_RCR_PROM_MASK; 
}

/* disableBroadcastReception():
 * Referring to the description of the MACRCTL register in
 * the datasheet, I don't think this device allows broadcast
 * reception to be disabled.
 */
void
disableBroadcastReception(void)
{
	ENET_RCR |= ENET_RCR_BC_REJ_MASK;
}

/* enableBroadcastReception():
 * Referring to the description of the MACRCTL register in
 * the datasheet, I think this device has broadcast reception
 * enabled by default (with no option to disable it).
 */
void
enableBroadcastReception(void)
{
	ENET_RCR &= ~ENET_RCR_BC_REJ_MASK; 
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	printf("Self-test not yet available for this driver.\n");
	//ENET_RCR |= ENET_RCR_LOOP_MASK;
	return(1);
}

static void
showBd(enet_bd_struct_t *bdp)
{
	printf("bd@0x%lx:\n",(long)bdp);
	printf(" len:  %d\n",htons((uint16_t)(bdp->length)));
	printf(" stat: x%04x\n",bdp->status);
	printf(" data: x%08lx\n",ecl((long)bdp->data));
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
	int i;
	printf("RX:\n");
	for(i=0;i<RXBUFTOT;i++)
		showBd(&rxbdtbl[i]);
	printf("TX:\n");
	for(i=0;i<TXBUFTOT;i++)
		showBd(&txbdtbl[i]);
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  The monitor will never need more than one packet at
 * a time; hence, we just keep giving out the same pointer.
 * Note that for the BF537, the first two bytes of the buffer are used
 * by the DMA engine for the length of the packet; so this function actually
 * hands the requestor a pointer incremented by two from the base of tx_data.
 */
static unsigned char *nxtsndbfr;

unsigned char *
getXmitBuffer(void)
{
	nxtsndbfr = txbuftbl[txidx];
	return(nxtsndbfr);
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 *
 * This driver uses register based DMA for ethernet transmit simply
 * because uMon only needs one transmit buffer; hence, there's really
 * no need for the added complexity of the descriptor model.
 * Note that DMA_2 is for ethernet transmit.
 */
int
sendBuffer(int length)
{
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)nxtsndbfr,length,ETHER_OUTGOING);
#endif

	txbdtbl[txidx].status |= TX_BD_TC | TX_BD_R | TX_BD_L;
	txbdtbl[txidx].length = htons(length);
	txbdtbl[txidx].data = (uint8_t *)htonl((ulong)txbuftbl[txidx]);
#ifdef ENHANCED_BD
	txbdtbl[txidx].bdu = 0x00000000;
	txbdtbl[txidx].ebd_status = TX_BD_INT | TX_BD_TS;
#endif
	ENET_TDAR = ENET_TDAR_TDAR_MASK;

	if (++txidx >= TXBUFTOT)
		txidx = 0;

	EtherXFRAMECnt++;
	return(0);
}

/* DisableEtherdev():
 */
void
DisableEtherdev(void)
{
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

#define PHY_STAT	1
#define PHY_STAT_LINKUP (1<<2)

int
linkpoll(void)
{
	char linkstate;
	uint16_t stat;

	if (phy_read(PHY_STAT,&stat) == -1)
		return(0);

	linkstate = (stat & PHY_STAT_LINKUP ? 1:0);

	if (linkstate != lastlinkstate) {
		if (linkstate) {
			// Link transitioned from down to up, so re-init...
			enreset();
			eninit();
		}
		printf("Link %s\n",linkstate ? "up":"dwn");
		lastlinkstate = linkstate;
	}
	return(lastlinkstate);
}

int
polletherdev(void)
{
	int pktcnt;
	uint16_t stat;
	enet_bd_struct_t *rbd;

	pktcnt = 0;

	if (linkpoll()) {
		rbd = &rxbdtbl[rxidx];
		if (((long)rbd & 0xf) != 0) {
			printf("misaligned bd!\n");
			return(0);
		}
		if ((ENET_RDAR & ENET_RDAR_RDAR_MASK) == 0) {
			// If we're here then all our RX buffers are filled, so we
			// need to cleanup.  Without this cleanup I found that the
			// receive side of the MAC basically shuts down (no more
			// packets received).
			enreset();
			eninit();
			return(0);
		}
		stat = rbd->status;
		if ((stat & RX_BD_E) == 0) {
			if (++rxidx >= RXBUFTOT)
				rxidx = 0;
			processPACKET((struct ether_header *)htonl((long)rbd->data),htons(rbd->length));
			rbd->status &= RX_BD_W;	// clear all bits except WRAP
			rbd->status |= RX_BD_E;	// set empty bit
			rbd->length = 0;
#ifdef ENHANCED_BD
			rbd->bdu = 0x00000000;
			rbd->ebd_status = RX_BD_INT;
#endif               
			pktcnt++;
		}
	}
	return (pktcnt);
}

int 
phy_read(int reg_addr, uint16_t *data)
{
	struct elapsed_tmr tmr;

	/* Clear the MII interrupt bit */
	ENET_EIR = ENET_EIR_MII_MASK;

	/* Initiatate the MII Management read */
	ENET_MMFR = ENET_MMFR_ST(0x01) | ENET_MMFR_OP(0x2)
		| ENET_MMFR_PA(PHYADDR) | ENET_MMFR_RA(reg_addr)
		| ENET_MMFR_TA(0x02);

	/* Poll for the MII interrupt (interrupt should be masked) */
	startElapsedTimer(&tmr,100);
	while((ENET_EIR & ENET_EIR_MII_MASK) == 0) {
		if (msecElapsed(&tmr)) {
			printf("phy_read timeout\n");
			return(-1);
		}
	}
    
	/* Clear the MII interrupt bit */
	ENET_EIR = ENET_EIR_MII_MASK;

	*data = ENET_MMFR & 0x0000FFFF;

	return(0);
}

int
phy_write(int reg_addr, uint16_t data)
{
	struct elapsed_tmr tmr;

	/* Clear the MII interrupt bit */
	ENET_EIR = ENET_EIR_MII_MASK;

	/* Initiatate the MII Management write */
	ENET_MMFR = ENET_MMFR_ST(0x01) | ENET_MMFR_OP(0x01)
		| ENET_MMFR_PA(PHYADDR) | ENET_MMFR_RA(reg_addr)
		| ENET_MMFR_TA(0x02) | ENET_MMFR_DATA(data);

	/* Poll for the MII interrupt (interrupt should be masked) */
	startElapsedTimer(&tmr,100);
	while((ENET_EIR & ENET_EIR_MII_MASK) == 0) {
		if (msecElapsed(&tmr)) {
			printf("phy_write timeout\n");
			return(-1);
		}
	}

	/* Clear the MII interrupt bit */
	ENET_EIR = ENET_EIR_MII_MASK;

	return(0);
}

#endif
